#!/usr/bin/env python3

import argparse

from atlas.lib.executor import executor_silent
from atlas.lib.screen import printer
from atlas.lib.executor import executor 
from atlas.lib.linker_loader import linker


# https://sumit-ghosh.com/articles/parsing-dictionary-key-value-pairs-kwargs-argparse-python/
class parse_params(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, dict())
        for value in values:
            key, value = value.split('=')
            getattr(namespace, self.dest)[key] = value


class ATLAS:

    def __init__(self, file_name: str, param: dict={}, printer: object=None) -> None:
        self.file_name = file_name
        self.param = param
        self.printer = printer

        self.linked = linker(self.file_name, self.param)

    def execute(self) -> bool:
        result = False

        if self.linked != None:
            if self.printer != None:
                result = executor(self.linked, self.printer)
            else:
                result = executor_silent(self.linked)

        return result

    @property
    def chain(self: object) -> object:
        return self.linked.CHAIN


def main() -> bool:
    s_printer = printer()
    epilog = """Example Usage:

  atlas -a HelloWorld.atl
  atlas -a some.atl --param foo=bar
    """
    argumentParser = argparse.ArgumentParser(
        "ATLAS",
        epilog=epilog,
        formatter_class=argparse.RawDescriptionHelpFormatter)
    
    argumentParser.add_argument(
        '-a', '--atl-file',
        nargs=1,
        required=True,
        help='ATLAS rule file'
    )
    argumentParser.add_argument(
        '-p', '--param',
        nargs='*',
        required=False,
        help='Parameters to use inside the rule, key=value',
        action=parse_params
    )
    args = argumentParser.parse_args()
    
    file_name = args.atl_file[0]

    atlas = ATLAS(file_name, args.param, s_printer)
    atlas.execute()

    return True


if __name__ == "__main__":
    main()


class UndefinedFuncValue(Exception):
    def __init__(self, func: str):
        super().__init__(f'{func} is missing in the core library.')


class UndefinedExpectValue(Exception):
    def __init__(self, expect: str):
        super().__init__(f'{expect} is missing in the expect library.')


class ParamKeyMissing(Exception):
    def __init__(self, key: str):
        super().__init__(f'{key} is missing in command-line argument param.')


class SpecialKeyUsage(Exception):
    def __init__(self, key: str):
        super().__init__(f'{key}')


class DuplicateKeyUsage(Exception):
    def __init__(self, key: str):
        super().__init__(f'{key}')


class RunKeyUsage(Exception):
    def __init__(self):
        super().__init__()


class ChainSectionMissing(Exception):
    def __init__(self):
        super().__init__(f'The rule doesn\'t have chain section.')


class ChainSectionEmpty(Exception):
    def __init__(self):
        super().__init__(f'The rule\'s chain section is empty.')


class FuncMissing(Exception):
    def __init__(self, subchain: str):
        super().__init__(f'The {subchain} must have func key')


class ExpectMissing(Exception):
    def __init__(self, subchain: str):
        super().__init__(f'The {subchain} must have expect key')


class ScriptsBase64ValueError(Exception):
    def __init__(self, script: str):
        super().__init__(f'The {script}\'s value raise an Base64 exception.')





import re
import os
import yaml
import base64

from glob import glob
from pprint import pprint
from collections import OrderedDict

from . import core, expect, error


LEXICON = {
    'headers': [
        'meta',
        'modules',
        'scripts',
        'chain'
    ],
    'words': [
        'input',
        'func',
        'output',
        'expect',
        'run'
    ]
}

ATL = {}
META = {}
CHAIN = []
MODULES = {}
MODULE_NAMES = set()
SCRIPTS = {}

SUBCHAIN_KEYS = []

class node:

    input = None
    func = None
    func_str = ""
    expect = None
    expect_str = []
    output = None
    output_str = ""

    def __init__(self: object, input: list, func: str, expect: str = "", output: str = "") -> None:
        self.input = input
        self.func_str = func
        self.expect_str = expect
        self.output_str = output
        self.expect = []


class linker:

    file_name = ""
    base_name = ""
    atl_yaml = yaml
    META = {}
    CHAIN = []
    MODULES = {}
    SCRIPTS = {}

    def __init__(self: object, file_name: str, param: dict={}) -> None:
        global ATL
        self.file_name = file_name
        self.param = param

        temp_file_name, file_extension = os.path.splitext(self.file_name)
        self.base_name = os.path.basename(temp_file_name)

        self.atl_yaml = self.link(self.file_name)
        
        if self.atl_yaml != None:
            temp = ATL
            ATL = {}
            ATL[self.base_name] = temp
            ATL[self.base_name] = {}
            ATL[self.base_name]['chain'] = self.CHAIN
            ATL[self.base_name]['scripts'] = self.SCRIPTS

            for i in self.MODULES:
                ATL[self.base_name][i] = self.MODULES[i]
            
            self.loader()

            


    def link(self: object, file_name: str) -> yaml:
        result = None
        atl = ""
        atl_loaded = None

        if not os.path.isfile(file_name):
            return result
        
        with open(file_name, 'r') as file:
            atl = file.read()

        try:
            atl_loaded = yaml.load(atl, Loader=yaml.Loader)
        except Exception as e:
            print(e)
            return result

        if len(atl_loaded.get('chain', [])) > 0 and self.lexical_analysis(atl_loaded):
            result = atl_loaded

        return result

    def lexical_analysis(self: object, atl: list) -> bool:
        result = False
        is_chain = False
        
        for i in atl:
            if i not in LEXICON['headers']:
                return result
            if i == 'chain':
                is_chain = True

        if not is_chain:
            return result

        if not self.lexical_analysis_meta(atl.get('meta', [])) or \
           not self.lexical_analysis_scripts(atl.get('scripts', [])) or \
           not self.lexical_analysis_modules(atl.get('modules', [])) or \
           not self.lexical_analysis_chain(atl['chain']):
            return result
        
        result = True

        return result


    def lexical_analysis_chain(self: object, chain: list) -> bool:
        global SUBCHAIN_KEYS

        result = False
        temp = OrderedDict()
        input = []
        func = ""
        expect = []
        output = ""
        sub_input = []
        sub_func = ""
        sub_expect = []
        sub_output = ""
        sub_chain = []

        for i in chain:
            sub_chain = chain[i]
            input = []
            func = ""
            expect = []
            output = ""
            
            if i in SUBCHAIN_KEYS:
                raise(error.DuplicateKeyUsage(i))
            elif i in LEXICON['headers'] or \
                 i in LEXICON['words']:
                raise(error.SpecialKeyUsage(i))
            
            SUBCHAIN_KEYS.append(i)

            if type(chain[i]) is str and \
               chain[i][0] == '$' and \
               '.' in chain[i]:
                keys = chain[i][1:].split('.')
                value = self.MODULES
                
                try:
                    for j in keys:
                        value = value[j]
                    
                    if type(value) is OrderedDict and \
                       keys[1] == 'chain':
                        for j in value:
                            temp[j] = value[j]
                    else:
                        temp[i] = value
                except Exception as e:
                    print(e)
                    return result

            elif type(sub_chain) is dict:
                for j in sub_chain:
                    keys = []
                    value = ""

                    if j == 'input':
                        if type(sub_chain[j]) is str:
                            input.append(sub_chain[j])
                        elif type(sub_chain[j]) is list:
                            input += sub_chain[j]
                        else:
                            return result
                    elif j == 'func':
                        func = sub_chain[j]
                    elif j == 'expect':
                        if type(sub_chain[j]) is str:
                            expect.append(sub_chain[j])
                        elif type(sub_chain[j]) is list:
                            expect += sub_chain[j]
                        else:
                            return result
                    elif j == 'output':
                        output = sub_chain[j]
                    elif type(sub_chain[j]) is dict:
                        sub_input = []
                        sub_func = ""
                        sub_expect = []
                        sub_output = ""

                        if j in SUBCHAIN_KEYS:
                            raise(error.DuplicateKeyUsage(j))
                        elif j in LEXICON['headers'] or \
                             j in LEXICON['words']:
                            raise(error.DuplicateKeyUsage(j))
                        
                        SUBCHAIN_KEYS.append(j)

                        for k in sub_chain[j]:

                            if k == 'input':
                                if type(sub_chain[j][k]) is str:
                                    sub_input.append(sub_chain[j][k])
                                elif type(sub_chain[j][k]) is list:
                                    sub_input += sub_chain[j][k]
                                else:
                                    return result
                            elif k == 'func':
                                sub_func = sub_chain[j][k]
                            elif k == 'expect':
                                if type(sub_chain[j][k]) is str:
                                    sub_expect.append(sub_chain[j][k])
                                elif type(sub_chain[j][k]) is list:
                                    sub_expect += sub_chain[j][k]
                                else:
                                    return result
                            elif k == 'output':
                                sub_output = sub_chain[j][k]
                            else:
                                return result

                        # if (not len(sub_input) > 0 and not len(input)) or \
                        if sub_func == "" or \
                           (sub_expect == [] and expect == []):
                            return result
                        
                        if temp.get(i, {}) == {}:
                            temp[i] = OrderedDict()

                        if sub_expect == [] and \
                           expect != []:
                            sub_expect = expect
                        temp[i][j] = node(input + sub_input, sub_func, sub_expect, sub_output)

                        continue
                    else:
                        return result
                
                if func == "" and sub_func == "":
                    return result

                if temp.get(i, []) == []:
                    temp[i] = node(input, func, expect, output)
            else:
                return result

        result = True
        self.CHAIN = temp

        return result
        

    def lexical_analysis_scripts(self: object, scripts: list) -> bool:
        result = False
        
        if scripts == []:
            result = True
        else:
            for i in scripts:
                pre_result = False

                if type(i) is not str or type(scripts[i]) is not str:
                    return result
                elif i == 'run':
                    raise(error.RunKeyUsage)
                try:
                    script_bytes = bytes(scripts[i], 'ascii')

                    if base64.b64encode(base64.b64decode(script_bytes)) != script_bytes:
                        return result
                except Exception as e:
                    print(e)
                    return result

            result = True

        self.SCRIPTS = scripts

        return result


    def lexical_analysis_meta(self: object, meta: list) -> bool:
        result = False
        
        if meta == []:
            result = True
        else:
            for i in meta:
                if type(i) is not str or type(meta[i]) is not str:
                    return result

            result = True

        self.META = meta

        return result


    def lexical_analysis_modules(self: object, modules: list) -> bool:
        result = False
        
        if modules == []:
            result = True
        else:
            for i in modules:
                if type(i) is not str:
                    return result

            if not self.module_check(modules):
                return result

            result = True

        return result


    def module_check(self: object, modules: list) -> bool:
        result = False
        module_paths = []
        global MODULES
        global MODULE_NAMES

        if modules == []:
            result = True
        else:
            for i in modules:
                temp = []
                file_name = ""
                file_extension = ""

                try:
                    file_name, file_extension = os.path.splitext(i)
                except Exception as e:
                    print(e)
                    return result
                
                if file_extension == '':
                    i = i + '.atl'

                if '*' in i:
                    try:
                        temp = glob.glob(i)
                    except Exception as e:
                        print(e)
                        return result
                else:
                    temp.append(i)
                
                for j in temp:
                    module_linked = None
                    base_name = ""
                    if file_name not in MODULE_NAMES:
                        try:
                            module_linked = linker(j)
                            if module_linked == None:
                                return result

                            base_name = os.path.basename(file_name)
                            self.MODULES[base_name] = {}
                            self.MODULES[base_name]['chain'] = module_linked.CHAIN
                            self.MODULES[base_name]['scripts'] = module_linked.SCRIPTS
                            MODULE_NAMES.add(file_name)
                        except Exception as e:
                            print(e)
                            return result

        result = True

        return result


    def loader(self: object) -> bool:
        result = False
        global ATL
        global SUBCHAIN_KEYS

        for i in self.CHAIN:
            if type(self.CHAIN[i]) == OrderedDict:
                for j in self.CHAIN[i]:
                    try:
                        self.CHAIN[i][j].func = getattr(core, self.CHAIN[i][j].func_str)

                        if self.CHAIN[i][j].expect_str != []:
                            for k in self.CHAIN[i][j].expect_str:
                                self.CHAIN[i][j].expect.append(getattr(expect, k))
                    except Exception as e:
                        print(e)
                        return result
                    
                    for k in range(0, len(self.CHAIN[i][j].input)):
                        if self.CHAIN[i][j].input[k][0] == '$':
                            keys = self.CHAIN[i][j].input[k][1:].split('.')
                            if keys[0] == 'param':
                                try:
                                    self.CHAIN[i][j].input[k] = self.param[keys[1]]
                                except:
                                    raise(error.ParamKeyMissing(keys[1]))
                            else:
                                value = ATL[self.base_name]

                                try:
                                    for z in keys:
                                        value = value[z]
                                    
                                    if re.search(r'^.*scripts\.[a-zA-Z0-9_]+$', self.CHAIN[i][j].input[k][1:]):
                                        self.CHAIN[i][j].input[k] = (value, keys[len(keys) - 1])
                                    else:
                                        self.CHAIN[i][j].input[k] = value
                                except Exception as e:
                                    if z not in SUBCHAIN_KEYS:
                                        return result
                                    continue
                            

            else:
                try:
                    self.CHAIN[i].func = getattr(core, self.CHAIN[i].func_str)

                    if self.CHAIN[i].expect_str != []:
                        for j in self.CHAIN[i].expect_str:
                            self.CHAIN[i].expect.append(getattr(expect, j))
                except Exception as e:
                    print(e)
                    return result

                for j in range(0, len(self.CHAIN[i].input)):
                    if self.CHAIN[i].input[j][0] == '$':
                        keys = self.CHAIN[i].input[j][1:].split('.')
                        if keys[0] == 'param':
                            try:
                                self.CHAIN[i].input[j] = self.param[keys[1]]
                            except:
                                raise(error.ParamKeyMissing(keys[1]))
                        else:
                            value = ATL[self.base_name]

                            try:
                                for z in keys:
                                    value = value[z]
                                if re.search(r'^.*scripts\.[a-zA-Z0-9_]+$', self.CHAIN[i].input[j][1:]):
                                    self.CHAIN[i].input[j] = (value, keys[len(keys) - 1])
                                else:
                                    self.CHAIN[i].input[j] = value
                            except Exception as e:
                                if z not in SUBCHAIN_KEYS:
                                    return result
                                continue
        result = True

        return result


import time
import threading
import json
import io

from typing import Dict, List, final
from contextlib import redirect_stdout, redirect_stderr
from collections import OrderedDict

from . import linker_loader


def node_to_dict(node: object, stdout: str, stderr: str) -> Dict:
    result = {
        'func': '',
        'input': [],
        'expect': [],
        'output': None
    }

    try:
        result['func'] = node.func_str
        result['expect'] = node.expect_str
        
        for i in node.input:
            if type(i) == str:
                result['input'].append(i[:50])
            elif type(i) == int:
                result['input'].append(int(str(i)[:50]))
            elif type(i) == bytes:
                result['input'].append(str(i[:50]))
            elif type(i) == bool:
                result['input'].append(i)
            else:
                result['input'].append(str(type(i)))

        if type(node.output) == str:
            result['output'] = node.output[:50]
        elif type(node.output) == int:
            result['output'] = int(str(node.output)[:50])
        elif type(node.output) == bytes:
            result['output'] = str(node.output[:50])
        elif type(node.output) == bool:
            result['output'] = node.output
        else:
            result['output'] = str(type(node.output))
        

        result['stdout'] = stdout[:200]
        result['stderr'] = stderr[:200]
    except Exception as e:
        print(e)
    
    return result



def executor(linked: object, s_printer: object) -> bool:
    result = False

    for i in linked.CHAIN:
        res = None
        if type(linked.CHAIN[i]) is OrderedDict:
            print('\033[4m' + '\033[1m' + f'[+] {i}' + '\033[0m')
            for j in linked.CHAIN[i]:
                res = None
                positive = False

                thread1 = threading.Thread(
                    target=s_printer.status,
                    args=(j, 1,),
                    daemon=True)
                thread1.start()

                for k in range(0, len(linked.CHAIN[i][j].input)):
                    if linked.CHAIN[i][j].input[k][0] == '$':
                        keys = linked.CHAIN[i][j].input[k][1:].split('.')
                        temp = linked.CHAIN

                        try:
                            for z in range(len(keys)):
                                temp = temp[keys[z]]
                                if type(temp) is linker_loader.node and\
                                   type(temp.output) is dict and\
                                   len(keys) - 1 > z:
                                    temp = temp.output
                            if type(temp) is linker_loader.node:
                                linked.CHAIN[i].input[k] = temp.output
                            else:
                                linked.CHAIN[i].input[k] = temp
                        except Exception as e:
                            print(e)
                            return result
                

                with redirect_stdout(io.StringIO()) as stdout, \
                     redirect_stderr(io.StringIO()) as stderr:
                    res = linked.CHAIN[i][j].func(*linked.CHAIN[i][j].input)

                if linked.CHAIN[i][j].expect != []:
                    for k in range(len(linked.CHAIN[i][j].expect)):
                        if not linked.CHAIN[i][j].expect[k](res):
                            s_printer.return_code = 99
                            thread1.join()
                            node_dict = node_to_dict(linked.CHAIN[i][j], stdout.getvalue(), stderr.getvalue())
                            print(json.dumps(node_dict, indent=8))
                            print()
                            break
                        elif k == len(linked.CHAIN[i][j].expect) - 1:
                            positive = True

                            s_printer.return_code = 99
                            thread1.join()
                            node_dict = node_to_dict(linked.CHAIN[i][j], stdout.getvalue(), stderr.getvalue())
                            print(json.dumps(node_dict, indent=8))
                            print()
                            
                            linked.CHAIN[i][j].output = res
                            linked.CHAIN[i] = linked.CHAIN[i][j]
                    if positive:
                        break
                elif res == True:
                    positive = True
                    linked.CHAIN[i] = linked.CHAIN[i][j]
                    
                    s_printer.return_code = 99
                    thread1.join()
                    node_dict = node_to_dict(linked.CHAIN[i][j], stdout.getvalue(), stderr.getvalue())
                    print(json.dumps(node_dict, indent=8))
                    print()

                    break
            
            if not positive:
                return result
        else:
            thread1 = threading.Thread(
                target=s_printer.status,
                args=(i,),
                daemon=True)
            thread1.start()
            for k in range(0, len(linked.CHAIN[i].input)):
                if linked.CHAIN[i].input[k][0] == '$':
                    keys = linked.CHAIN[i].input[k][1:].split('.')
                    temp = linked.CHAIN

                    try:
                        for z in range(len(keys)):
                            temp = temp[keys[z]]
                            if type(temp) is linker_loader.node and\
                               type(temp.output) is dict and\
                               len(keys) - 1 > z:
                                temp = temp.output
                        if type(temp) is linker_loader.node:
                            linked.CHAIN[i].input[k] = temp.output
                        else:
                            linked.CHAIN[i].input[k] = temp
                    except Exception as e:
                        print(e)
                        return result


            with redirect_stdout(io.StringIO()) as stdout, \
                 redirect_stderr(io.StringIO()) as stderr:
                res = linked.CHAIN[i].func(*linked.CHAIN[i].input)

            if linked.CHAIN[i].expect != []:
                for j in range(len(linked.CHAIN[i].expect)):
                    if not linked.CHAIN[i].expect[j](res):
                        return result
                    elif j == len(linked.CHAIN[i].expect) - 1:
                        linked.CHAIN[i].output = res
            else:
                linked.CHAIN[i].output = res
            
            if stderr.getvalue():
                s_printer.return_code = 99
                thread1.join()
                node_dict = node_to_dict(linked.CHAIN[i], stdout.getvalue(), stderr.getvalue())
                print(json.dumps(node_dict, indent=4))
                print()

                return result
            else:
                s_printer.return_code = 0
                thread1.join()
                node_dict = node_to_dict(linked.CHAIN[i], stdout.getvalue(), stderr.getvalue())
                print(json.dumps(node_dict, indent=4))
                print()
            

    result = True

    return result


def executor_silent(linked: object) -> bool:
    result = False

    for i in linked.CHAIN:
        res = None
        if type(linked.CHAIN[i]) is OrderedDict:
            for j in linked.CHAIN[i]:
                res = None
                positive = False

                for k in range(0, len(linked.CHAIN[i][j].input)):
                    if linked.CHAIN[i][j].input[k][0] == '$':
                        keys = linked.CHAIN[i][j].input[k][1:].split('.')
                        temp = linked.CHAIN

                        try:
                            for z in range(len(keys)):
                                temp = temp[keys[z]]
                                if type(temp) is linker_loader.node and\
                                   type(temp.output) is dict and\
                                   len(keys) - 1 > z:
                                    temp = temp.output
                            if type(temp) is linker_loader.node:
                                linked.CHAIN[i].input[k] = temp.output
                            else:
                                linked.CHAIN[i].input[k] = temp
                        except Exception as e:
                            print(e)
                            return result
                

                with redirect_stdout(io.StringIO()) as stdout, \
                     redirect_stderr(io.StringIO()) as stderr:
                    res = linked.CHAIN[i][j].func(*linked.CHAIN[i][j].input)

                if linked.CHAIN[i][j].expect != []:
                    for k in range(len(linked.CHAIN[i][j].expect)):
                        if not linked.CHAIN[i][j].expect[k](res):
                            break
                        elif k == len(linked.CHAIN[i][j].expect) - 1:
                            positive = True
                            
                            linked.CHAIN[i][j].output = res
                            linked.CHAIN[i] = linked.CHAIN[i][j]
                    if positive:
                        break
                elif res == True:
                    positive = True
                    linked.CHAIN[i] = linked.CHAIN[i][j]

                    break
            
            if not positive:
                return result
        else:
            for k in range(0, len(linked.CHAIN[i].input)):
                if linked.CHAIN[i].input[k][0] == '$':
                    keys = linked.CHAIN[i].input[k][1:].split('.')
                    temp = linked.CHAIN

                    try:
                        for z in range(len(keys)):
                            temp = temp[keys[z]]
                            if type(temp) is linker_loader.node and\
                               type(temp.output) is dict and\
                               len(keys) - 1 > z:
                                temp = temp.output
                        if type(temp) is linker_loader.node:
                            linked.CHAIN[i].input[k] = temp.output
                        else:
                            linked.CHAIN[i].input[k] = temp
                    except Exception as e:
                        print(e)
                        return result


            with redirect_stdout(io.StringIO()) as stdout, \
                 redirect_stderr(io.StringIO()) as stderr:
                res = linked.CHAIN[i].func(*linked.CHAIN[i].input)

            if linked.CHAIN[i].expect != []:
                for j in range(len(linked.CHAIN[i].expect)):
                    if not linked.CHAIN[i].expect[j](res):
                        return result
                    elif j == len(linked.CHAIN[i].expect) - 1:
                        linked.CHAIN[i].output = res
            else:
                linked.CHAIN[i].output = res
            
            if stderr.getvalue():
                return result

    result = True

    return result



# https://github.com/sulinx/remote_importer/blob/master/remote_importer.py
# http://wakandan.github.io/2018/06/07/python3-import-hook.html
# https://www.xorrior.com/In-Memory-Python-Imports/

import re
import sys
import types


MODULES = {}

class atl_importer():
    
    module_names = ""
    script = ""

    def __init__(self: object, *args) -> None:
        self.module_names = args
        self.path = None

    def find_module(self: object, fullname: str, path: str = None) -> object:
        global MODULES

        if re.search(r'^atl_.+?$', fullname) and \
           MODULES.get(fullname, '') != '':
            return self

        return None

    def load_module(self: object, name: str):
        global MODULES

        if name in sys.modules:
            return sys.modules[name]

        source = MODULES[name]
        code = compile(source, 'test', 'exec')
        mod = sys.modules.get(name)

        if mod is None:
            mod = sys.modules[name] = types.ModuleType(name)

        mod.__loader__ = self
        mod.__name__ = name
        exec(code, mod.__dict__)
        
        return mod

sys.meta_path.insert(-1, atl_importer())


import re
import sys
import base64
import hashlib
import platform
import requests

from subprocess import Popen, TimeoutExpired, PIPE

from typing import Dict, List

from . import atl_importer

POWERSHELL_DICT = {
                'Linux': 'pwsh',
                'Darwin': 'pwsh',
                'Java': None,
                'Windows': 'powershell.exe'
}

POWERSHELL_NAME = ""

try:
    plat = platform.system()
    POWERSHELL_NAME = POWERSHELL_DICT[plat]
except Exception as e:
    print(e)


def reverse(data: any) -> any:
    return data[::-1]

def download_from_remote_server(addr: str) -> bytes:
    result = b''

    try:
        response = requests.get(url=addr)
        result = response.content
    except Exception as e:
        sys.stderr.write(str(e))
        
    return result


def powershell_executor(script_name: tuple, *args) -> any:
    global POWERSHELL_NAME
    result = None
    script, name = script_name
    args_list = []

    with Popen([POWERSHELL_NAME, "-Command", r"-"], stdin=PIPE, stdout=PIPE, stderr=PIPE, bufsize=8192, universal_newlines=True) as proc:

        for i in range(int(len(script) / 10000) + 1):
            proc.stdin.write('$script=$script+"{}"\r\n'.format(script[i * 10000:(i + 1) * 10000]))
            proc.stdin.flush()

        proc.stdin.write('\
                $decoded_script=[System.Text.Encoding]::UTF8.GetString([System.Convert]::FromBase64String($script));\
                $ScriptBlock=[ScriptBlock]::Create($decoded_script);\
                . $ScriptBlock;\r\n\
            ')
        proc.stdin.flush()
        
        for i in range(len(args)):
            temp = ""
            if type(args[i]) is str:
                temp = args[i]
                for j in range(0, int(len(temp) / 10000) + 1):
                    proc.stdin.write('$arg{}=$arg{}+"{}";\r\n'.format(i + 1, i + 1, temp[j * 10000:(j + 1) * 10000]))
                    proc.stdin.flush()
                    #print('$arg{}=$arg{}+"{}"\r\n'.format(i + 1, i + 1, temp[j * 10000:(j + 1) * 10000]))
            elif type(args[i]) is bytes:
                temp = base64.b64encode(args[i]).decode()
                for j in range(0, int(len(temp) / 10000) + 1):
                    proc.stdin.write('$encoded_arg{}=$encoded_arg{}+"{}";\r\n'.format(i + 1, i + 1, temp[j * 10000:(j + 1) * 10000]))
                    proc.stdin.flush()
                
                proc.stdin.write('$arg{}=([System.Convert]::FromBase64String($encoded_arg{}));'.format(i + 1, i + 1))
                proc.stdin.flush()

            args_list.append("$arg{}".format(i + 1))
        proc.stdin.write('\
                try {{\
                    $result={0} {1};\
                }}\
                catch {{\
                    $result=run {1};\
                }}\
                \
                if ( $result.gettype().Name -eq "Byte[]" )\
                {{\
                    $result_encoded=[System.Convert]::ToBase64String($result);\
                }}\
                else\
                {{\
                    $result_encoded=[System.Convert]::ToBase64String([System.Text.Encoding]::UTF8.GetBytes($result));\
                }}\
                $temp_file=New-TemporaryFile;\
                $return_file_path=Join-Path -Path $temp_file.DirectoryName -ChildPath $temp_file.Name;\
                Set-Content $return_file_path -Value $result_encoded -Encoding Ascii;\
                $return_file_path\r\n\
            '.format(name, " ".join(args_list)))

        # Because the PIPE's buffer filled up with data bigger than 64k, can't get the stdout directly.
        # Then decided to get it via tempfile.
        output, _ = proc.communicate()
        try:
            if proc.poll() != None:
                proc.kill()
            output = re.sub('(\n)|(\r)|(\\x1b\[\?1(h|l))', '', output)
            result = base64.b64decode(file_read_bin(output))
        except TimeoutExpired:
            proc.kill()
        except Exception as e:
            proc.kill()
            sys.stderr.write(str(e))
            return result
    
    return result

def file_read_bin(path: str) -> bytes:
    result = b''

    try:
        with open(path, 'rb') as file:
            result = file.read()
    except Exception as e:
        sys.stderr.write(str(e))

    return result

def file_read_utf8(path: str) -> str:
    result = ''

    try:
        with open(path, 'r', encoding="utf8") as file:
            result = file.read()
    except Exception as e:
        sys.stderr.write(str(e))

    return result
    

def save_file_bytes(data: any, prefix: str = '') -> str:
    result = ""
    sha256 = get_sha256(data)

    try:
        with open("{}{}.bin".format(prefix, sha256), 'wb') as file:
            if type(data) is bytes:
                file.write(data)
            elif type(data) is str:
                file.write(data.encode())
            else:
                return result
    except Exception as e:
        sys.stderr.write(str(e))
        return result
    
    result = "{}{}.bin".format(prefix, sha256)

    return result


def save_file_arr(arr: list, prefix: str = 'output') -> bool:
    result = False

    try:
        for i in range(0, len(arr)):
            with open("{}_{}".format(prefix, i), 'wb') as file:
                if type(arr[i]) is bytes:
                    file.write(arr[i])
                elif type(arr[i]) is str:
                    file.write(arr[i].encode())
                else:
                    return result
    except Exception as e:
        sys.stderr.write(str(e))
        return result
    
    result = True

    return result


def python_executor(script_tuple: tuple, *args) -> any:
    result = None
    decoded_script = ""
    script, script_name = script_tuple
    name = 'atl_' + script_name

    try:
        decoded_script = base64.b64decode(script)
        
        if atl_importer.MODULES.get(name, '') == '':
            atl_importer.MODULES[name] = decoded_script
            imported = __import__(name)
        else:
            imported = sys.modules[name]

        try:
            func = getattr(imported, script_name)
        except AttributeError:
            func = getattr(imported, 'run')

        result = func(*args)
    except Exception as e:
        sys.stderr.write(str(e))

    return result


def printer(*args) -> bool:

    print(" ".join(args))
    return True


def hello_world() -> None:
    print("Hello World, ATLAS.")

    return True

def get_sha256(data: bytes) -> str:
    sha256 = ''
    
    try:
        sha256 = hashlib.sha256(bytearray(data)).hexdigest()
    except Exception as e:
        sys.stderr.write(str(e))

    return sha256

def bytes_to_str_utf8(data: bytes=b'') -> str:
    try:
        return data.decode('utf8')
    except Exception as e:
        sys.stderr.write(str(e))
        return ''

    


def is_not_none(data: any) -> bool:
    return not data == None

def is_not_empty_list(data: list) -> bool:
    if type(data) is not list:
        return False
    
    return len(data)

def is_not_empty_dict(data: dict) -> bool:
    if type(data) is not dict:
        return False
    
    return len(data)

def is_not_empty_str(data: str) -> bool:
    if type(data) is not str:
        return False
    
    return data == ""

try:
    import magic

    def is_pe(data: bytes) -> bool:
        try:
            data_magic = magic.from_buffer(data, mime = True)
        except Exception as e:
            return False

        return data_magic == 'application/x-dosexec'
except:
    pass

def is_true(boolean: bool) -> bool:
    return True == boolean



import os
import json
import time
import pprint

from typing import Dict, List

class printer:
    return_code = -1
    rows = -1
    columns = -1
    search = 0
    dTot = 0
    dCur = 0
    plat = None

    atlas_banner = \
        """             .:                                 ;.           
           ;xkl'                              .,ox;          
         .cOd,kK0k                          0xcl.cxl.        
        .lOl. .:dOko;..                 .':okd:. doko.       
        ckc.    .'cdkOxol:,,'.....',;:ldxkdc'.    dx0l.      
       ,ko.         .':loxkkkkkkkkkkxdl:,.        .k0k;      
      .lx'                x0KKKK00k:               ,xoc.     
      .dl                    l33t                  ,l.c.     
      .d:                     ''                     .;'     
      .o;                                            o,.     
      .::.                                           .0.     
      .lxl.                                          .c.     
      .lOl.                                         ,;K.     
       .okc.                                       lO,x      
        .oOxc.                .;cddl'             .oOl.      
         .l0KOo.             .d00KKKk,           .lx:.       
          .l0K0xclooc;;:cc:,;d0KKKK0x'       .'cdOd'         
           .ck0KKKKKK00KKKK00KKKKK0d,.....  ,x0KOl.          
             .:ldkO0000KKKKKKKK00K0OxxxkkkdokK0k:.           
                 ...';dxxOKKK0KK0KKKK000KKK0Oxc.             
                     .lOxdxO0KKK000Od;;:::;,..               
                      'xK0OO0KKK0kl'                         
                       c0KKKKKKKKO:                          
                       ;OKKKKKKK0o.                          
                       cOKKKKKKKOc..',,;;;,'.                
                      cO0KKKKKKK0OkO0000K000kl.              
                     .xKKKKKKKKKKKKKKKKKKKKKKx.              
                     .dKKKKKKKK0KKK00Oxk0KKKO:               
                      'x0KKKKK00kl:;,';x0KK0l.               
                       'x0KKKKKKx.   .oKKK0l.                
                        'd0KKKKK0:   .xKK0l.                 
                         'x0KKKK0l.  ,kKOc.                  
            .:c;.  ..,cldxO0KKKK0l  .d00l.                   
        .,:ok0K0OxxkO0KK0000KK00x' .o0K0o;..                 
        .,;;;;;;;;;;;;;;;;;;;;,..  .clllll:,.                """


    def __init__(self, 
                 is_banner: bool=True) -> None:
        if is_banner:
            print('\033[2J\033[1;1H')
            print(self.atlas_banner)
            print()
            print('\033[1m' + "        ATLAS - Malware Analysis Description" + '\033[0m')
            print()

    # https://stackoverflow.com/questions/287871/how-do-i-print-colored-text-to-the-terminal
    def status(self, 
               subchain: str,
               tab_count: int=0) -> None:

        while self.return_code == -1:
            print('\t' * tab_count + f"[|] {subchain}" + '\033[0m', end='\r')
            time.sleep(0.1)
            print('\t' * tab_count + f"[/] {subchain}" + '\033[0m', end='\r')
            time.sleep(0.1)
            print('\t' * tab_count + f"[-] {subchain}" + '\033[0m', end='\r')
            time.sleep(0.1)
            print('\t' * tab_count + f"[\\] {subchain}" + '\033[0m', end='\r')
            time.sleep(0.1)
            print('\t' * tab_count + f"[|] {subchain}" + '\033[0m', end='\r')
            time.sleep(0.1)
        if self.return_code == 0:
            print('\t' * tab_count + '\033[92m' + f"[+] " + subchain + '\033[0m')
        else:
            print('\t' * tab_count + '\033[91m' + f"[+] " + subchain + '\033[0m')
        self.return_code = -1
